from os.path import abspath, dirname, join
import sys

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage.integrity_check  # noqa: E402


def test_ramp_from_h5file():
    """"
    The data for this test was created using:


    ```
    import numpy as np
    import qpimage

    size = 50
    # background phase image with a ramp
    bg = np.repeat(np.linspace(0, 1, size), size).reshape(size, size)
    bg = .6 * bg - .8 * bg.transpose() + .2
    # phase image with random noise
    rsobj = np.random.RandomState(47)
    phase = rsobj.rand(size, size) - .5 + bg

    # create QPImage instance
    with qpimage.QPImage(data=phase,
                         which_data="phase",
                         h5file="bg_ramp.h5") as qpi:
        qpi.compute_bg(which_data="phase",  # correct phase image
                       fit_offset="fit",  # use bg offset from ramp fit
                       fit_profile="ramp",  # perform 2D ramp fit
                       border_px=5,  # use 5 px border around image
                       )
    ```
    """
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpimage.integrity_check.check(h5file, checks=["background"])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
from os.path import abspath, dirname, join
import sys
import tempfile

import h5py

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402


def test_series_h5file():
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with h5py.File(tf, mode="a") as fd:
        qps = qpimage.QPSeries(h5file=fd)
        assert len(qps) == 0
    # cleanup
    try:
        os.remove(tf)
    except:
        pass


def test_series_from_list():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file)
    qpi2 = qpi1.copy()

    qps = qpimage.QPSeries(qpimage_list=[qpi1, qpi2])
    assert len(qps) == 2
    assert qps.get_qpimage(0) == qps.get_qpimage(1)


def test_series_from_h5file():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file)
    qpi2 = qpi1.copy()

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"
                          ):
        pass

    qps2 = qpimage.QPSeries(h5file=tf, h5mode="r")
    assert len(qps2) == 2
    assert qps2.get_qpimage(0) == qpi1
    # cleanup
    try:
        os.remove(tf)
    except:
        pass


def test_series_meta():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file, meta_data={"wavelength": 300e-9})
    assert qpi1.meta["wavelength"] == 300e-9
    qps = qpimage.QPSeries(qpimage_list=[qpi1], meta_data={
                           "wavelength": 400e-9})
    assert qps.get_qpimage(0).meta["wavelength"] == 400e-9


def test_series_error_meta():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file)
    qpi2 = qpi1.copy()

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"
                          ):
        pass

    try:
        qpimage.QPSeries(h5file=tf, h5mode="r",
                         meta_data={"wavelength": 550e-9})
    except ValueError:
        pass
    else:
        assert False, "`meta_data` and `h5mode=='r'`"
    # cleanup
    try:
        os.remove(tf)
    except:
        pass


def test_series_error_key():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file)
    qpi2 = qpi1.copy()

    qps = qpimage.QPSeries(qpimage_list=[qpi1, qpi2])
    try:
        qps.get_qpimage(2)
    except KeyError:
        pass
    else:
        assert False, "get index 2 when length is 2"


def test_series_error_file_is_qpimage():
    h5file = join(dirname(abspath(__file__)), "data/bg_ramp.h5")
    qpi1 = qpimage.QPImage(h5file=h5file)
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpi1.copy(h5file=tf):
        pass

    try:
        qpimage.QPSeries(qpimage_list=[qpi1], h5file=tf)
    except ValueError:
        pass
    else:
        assert False, "tf is a qpimage file"
    # cleanup
    try:
        os.remove(tf)
    except:
        pass


def test_series_error_noqpimage():
    try:
        qpimage.QPSeries(qpimage_list=["hans", 1])
    except ValueError:
        pass
    else:
        assert False, "qpimage list must contain QPImages"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import h5py

from .core import QPImage
from .meta import MetaDict


class QPSeries(object):
    _instances = 0

    def __init__(self, qpimage_list=[], meta_data={},
                 h5file=None, h5mode="a"):
        """Quantitative phase image series

        Parameters
        ----------
        qpimage_list: list of QPImage
            A list of instances of :py:class:`qpimage.QPImage`.
        meta_data: dict
            Meta data associated with the input data
            (see :py:class:`qpimage.VALID_META_KEYS`). This overrides
            the meta data of the QPImages in `qpimage_list` and, if
            `h5file` is given and `h5mode` is not "r", overrides
            the meta data in `h5file`.
        h5file: str, h5py.Group, h5py.File, or None
            A path to an hdf5 data file where all data is cached. If
            set to `None` (default), all data will be handled in
            memory using the "core" driver of the :mod:`h5py`'s
            :class:`h5py:File` class. If the file does not exist,
            it is created. If the file already exists, it is opened
            with the file mode defined by `hdf5_mode`. If this is
            an instance of h5py.Group or h5py.File, then this will
            be used to internally store all data. If `h5file` is given
            and `qpimage_list` is not empty, all QPImages in
            `qpimage_list` are appended to `h5file` in the given order.
        h5mode: str
            Valid file modes are (only applies if `h5file` is a path)
              - "r": Readonly, file must exist
              - "r+": Read/write, file must exist
              - "w": Create file, truncate if exists
              - "w-" or "x": Create file, fail if exists
              - "a": Read/write if exists, create otherwise (default)
        """
        if isinstance(h5file, h5py.Group):
            self.h5 = h5file
            self._do_h5_cleanup = False
        else:
            if h5file is None:
                h5kwargs = {"name": "qpseries{}.h5".format(
                                                        QPSeries._instances),
                            "driver": "core",
                            "backing_store": False,
                            "mode": "a"}
            else:
                h5kwargs = {"name": h5file,
                            "mode": h5mode}
            self.h5 = h5py.File(**h5kwargs)
            self._do_h5_cleanup = True
        QPSeries._instances += 1

        if meta_data and h5mode == "r":
            msg = "`h5mode` must not be 'r' if `meta_data` is given!"
            raise ValueError(msg)

        # make sure self.h5 is not itself a QPImage file
        if "phase" in self.h5 and "amplitude" in self.h5:
            raise ValueError(
                "`h5file` is a QPImage file, not a QPSeries file!")

        # Write QPimage data to h5 file
        for qpi in qpimage_list:
            self.add_qpimage(qpi)

        # Update meta data
        if meta_data:
            meta = MetaDict(meta_data)
            for ii in range(len(self)):
                qpii = self.get_qpimage(index=ii)
                for mk in meta:
                    qpii.h5.attrs[mk] = meta[mk]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._do_h5_cleanup:
            self.h5.flush()
            self.h5.close()

    def __len__(self):
        keys = list(self.h5.keys())
        keys = [kk for kk in keys if kk.startswith("qpi_")]
        return len(keys)

    def add_qpimage(self, qpimage):
        """Add a QPImage instance to the QPSeries

        Parameters
        ----------
        qpimage: QPImage
            The QPImage that is added to the series
        """
        if not isinstance(qpimage, QPImage):
            raise ValueError("`qpimage` must be instance of QPImage!")

        # determine number of qpimages
        num = len(self)
        # indices start at zero; do not add 1
        name = "qpi_{}".format(num)
        group = self.h5.create_group(name)
        qpimage.copy(h5file=group)

    def get_qpimage(self, index):
        """Return a single QPImage of the series

        Parameters
        ----------
        index: int
            Index of the qpimage
        """
        name = "qpi_{}".format(index)
        if name in self.h5:
            group = self.h5[name]
        else:
            msg = "Index {} not found for QPSeries of length {}".format(
                index,
                len(self))
            raise KeyError(msg)
        return QPImage(h5file=group)



VALID_META_KEYS = ["medium index",
                   "pixel size",
                   "time",
                   "wavelength",
                   ]

OTHER_META_KEYS = ["qpimage version",
                   ]

META_KEYS = VALID_META_KEYS + OTHER_META_KEYS


class MetaDataMissingError(BaseException):
    pass


class MetaDict(dict):
    """A dictionary containing meta data variables

    Valid keynames are given in the
    :py:const:`qpimage.meta.VALID_META_KEYS` variable.

    Methods
    -------
    __setitem__
    """

    def __init__(self, *args, **kwargs):
        super(MetaDict, self).__init__(*args, **kwargs)
        for key in self:
            if key not in META_KEYS:
                raise KeyError("Unknown meta variable: '{}'".format(key))

    def __setitem__(self, key, value):
        """Set a meta data variable

        The key must be a valid key defined in the
        :py:const:`qpimage.meta.VALID_META_KEYS` variable.
        """
        if key not in META_KEYS:
            raise KeyError("Unknown meta variable: '{}'".format(key))
        super(MetaDict, self).__setitem__(key, value)

    def __getitem__(self, *args, **kwargs):
        if args[0] not in self:
            msg = "No meta data was defined for {}! ".format(args[0]) \
                  + "Please make sure you passed the dictionary `meta_data` " \
                  + "when creating the QPImage instance."
            raise MetaDataMissingError(msg)
        return super(MetaDict, self).__getitem__(*args, **kwargs)



from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402
import qpimage.image_data  # noqa: E402


def test_set_bg():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    pha = qpi.pha
    clspha = qpi._pha
    assert np.all(pha == clspha.image)
    clspha.set_bg(None, key="data")
    assert not np.all(pha == clspha.image)


def test_get_bg():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    bgpha = qpi.bg_pha
    clspha = qpi._pha
    assert np.all(bgpha == clspha.bg)
    assert np.all(bgpha == clspha.get_bg(key=None))


def test_get_bg_error():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    clspha = qpi._pha
    try:
        clspha.get_bg(key=None, ret_attrs=True)
    except ValueError:
        pass
    else:
        assert False, "attrs of combined bg not supported"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from os.path import abspath, dirname
import sys

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402
import qpimage.meta  # noqa: E402


def test_meta():
    mm = qpimage.meta.MetaDict()
    mm["wavelength"] = 100e-9
    assert mm["wavelength"] == 100e-9


def test_meta_error():
    mm = qpimage.meta.MetaDict()
    try:
        mm["peter"] = "hans"
    except KeyError:
        pass
    else:
        assert False, "invalid key 'peter' should not work!"

    try:
        qpimage.meta.MetaDict({"peter2": "hans2"})
    except KeyError:
        pass
    else:
        assert False, "invalid key 'peter2' should not work!"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np


def find_sideband(data):
    """Find the side band position of a hologram

    The hologram is Fourier-transformed and the side band
    is determined by finding the maximum amplitude in
    Fourier space.

    Parameters
    ----------
    data : 2d ndarray
        hologram image


    Returns
    -------
    sb_loc : tuple of floats
        (x,y) coordinates of the side band in Fourier space
        measured from the center in the shifted Fourier transform.
    """
    # fourier transform
    fft = fourier2dpad(data)

    # Zero padded (larger)
    order = len(fft)

    # remove lower part
    minlo = max(int(np.ceil(order / 42)), 5)
    fft[order // 2 - minlo:] = 0
    # remove axis part
    center = int(order / 2)
    fft[center - 3:center + 3, :] = 0
    fft[:, center - 3:center + 3] = 0

    am = np.argmax(np.abs(fft))
    y = am % order
    x = (am - y) / order

    diskrad = int(order / 23)

    xv = np.arange(order).reshape(-1, 1)
    yv = xv.reshape(1, -1)

    fft[np.where((xv - x)**2 + (yv - y)**2 > diskrad**2)] = 0

    return (int(x - order / 2), int(y - order / 2))


def fourier2dpad(data):
    """Compute the 2D Fourier transform with zero padding"""
    # Zero padding to square image
    (N, M) = data.shape

    # zero padding
    order = np.int(max(64., 2**np.ceil(np.log(2 * max(N, M)) / np.log(2))))
    datapad = np.pad(data, ((0, order - N), (0, order - M)), mode="constant")

    # fourier transform
    fft = np.fft.fftshift(np.fft.fft2(datapad))

    return fft


def get_field(data, sb_loc=None, filt_size=None, filt_type="gauss"):
    """Computes phase and amplitude from a holographic image

    Parameters
    ----------
    data : 2d ndarray
        Hologram data
    sb_loc : tuple of floats (x0,y0)
        Coordinates of the side band in Fourier space.
    filt_size : float
        Radius of the filter in Fourier space in px.
        If set to None, the radius will be estimated using `sb_loc`.
    filt_type : float
        One of {"disk", "gauss"}

    x0 and y0 are center of the filter
    R is factor for "radius" of filter (sqrt(x0² + y0²)/np.pi)

    filter_type can be "disk" or "gauss"

    Notes
    -----
    `sb_loc` are in FT of zero-padded image, see `fourier2dpad`.
    """
    if sb_loc is None:
        sb_loc = find_sideband(data)

    x0, y0 = sb_loc
    n, m = data.shape

    # fourier transform
    fft = fourier2dpad(data)

    # Zero padded (larger)
    order = len(fft)

    center = order / 2 - .5
    x = np.linspace(-center, center, order, endpoint=False)
    xv = x.reshape(-1, 1)
    yv = x.reshape(1, -1)

    if filt_size is None:
        # Estimate filter size based on the distance to the origin
        filt_size = int(np.ceil(min(np.abs(sb_loc) / 2)))

    if filt_type == "gauss":
        # gaussian filter_type
        afilter = np.exp(-((xv - x0)**2 + (yv - y0)**2) / (filt_size**2))
        afilter /= np.max(afilter)
    elif filt_type == "disk":
        # disk fitlering
        afilter = ((xv - x0)**2 + (yv - y0)**2) < filt_size**2
    else:
        print("Unknown filter: {}".format(filt_type))
    # filter here
    fft_filt = afilter * fft

    # Shift image to zero
    shifted = np.roll(np.roll(fft_filt, -x0, axis=0), -y0, axis=1)

    # inverse fourier transform
    field = np.fft.ifft2(np.fft.ifftshift(shifted))[:n, :m]
    return field



from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402


def test_offset_mean_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=5)
    assert np.allclose(np.mean(qpi.pha), -0.0011984843060164104)


def test_offset_mode_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="mode",
                   fit_profile="offset",
                   border_px=5)
    # Not so good result because of poor statistics
    assert np.allclose(np.mean(qpi.pha), -0.044022731691440088)


def test_offset_gauss_simple():
    size = 200
    bg = 2.5
    rsobj = np.random.RandomState(47)
    data = rsobj.normal(loc=bg, scale=.1, size=(size, size))
    assert np.allclose(np.mean(data), 2.5003227280152887)
    qpi = qpimage.QPImage(data=data, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="gauss",
                   fit_profile="offset",
                   border_px=5)
    # This yields the best results, because the input noise
    # is gaussian.
    assert np.allclose(np.average(qpi.pha), -0.00039833492077054762)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import copy

import h5py
import numpy as np
from skimage.restoration import unwrap_phase

from . import holo
from .image_data import Amplitude, Phase
from .meta import MetaDict
from ._version import version as __version__

VALID_INPUT_DATA = ["field", "phase", "hologram",
                    "phase,amplitude", "phase,intensity"]


class QPImage(object):
    _instances = 0

    def __init__(self, data=None, bg_data=None, which_data="phase",
                 meta_data={}, h5file=None, h5mode="a"
                 ):
        """Quantitative phase image manipulation

        This class implements various tasks for quantitative phase
        imaging, including phase unwrapping, background correction,
        numerical focusing, and data export.

        Parameters
        ----------
        data: 2d ndarray (float or complex) or list
            The experimental data (see `which_data`)
        bg_data: 2d ndarray (float or complex), list, or `None`
            The background data (must be same type as `data`)
        which_data: str
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            "field", "phase", "phase,amplitude", or "phase,intensity",
            where the latter two require an indexable object with
            the phase data as first element.
        meta_data: dict
            Meta data associated with the input data.
            see :py:class:`qpimage.VALID_META_KEYS`
        h5file: str, h5py.Group, h5py.File, or None
            A path to an hdf5 data file where all data is cached. If
            set to `None` (default), all data will be handled in
            memory using the "core" driver of the :mod:`h5py`'s
            :class:`h5py:File` class. If the file does not exist,
            it is created. If the file already exists, it is opened
            with the file mode defined by `hdf5_mode`. If this is
            an instance of h5py.Group or h5py.File, then this will
            be used to internally store all data.
        h5mode: str
            Valid file modes are (only applies if `h5file` is a path)
              - "r": Readonly, file must exist
              - "r+": Read/write, file must exist
              - "w": Create file, truncate if exists
              - "w-" or "x": Create file, fail if exists
              - "a": Read/write if exists, create otherwise (default)
        """
        if isinstance(h5file, h5py.Group):
            self.h5 = h5file
            self._do_h5_cleanup = False
        else:
            if h5file is None:
                h5kwargs = {"name": "qpimage{}.h5".format(QPImage._instances),
                            "driver": "core",
                            "backing_store": False,
                            "mode": "a"}
            else:
                h5kwargs = {"name": h5file,
                            "mode": h5mode}
            self.h5 = h5py.File(**h5kwargs)
            self._do_h5_cleanup = True
        QPImage._instances += 1
        # set meta data
        meta = MetaDict(meta_data)
        for key in meta:
            self.h5.attrs[key] = meta[key]
        if "qpimage version" not in self.h5.attrs:
            self.h5.attrs["qpimage version"] = __version__
        # set data
        for group in ["amplitude", "phase"]:
            if group not in self.h5:
                self.h5.create_group(group)
        self._amp = Amplitude(self.h5["amplitude"])
        self._pha = Phase(self.h5["phase"])
        if data is not None:
            # compute phase and amplitude from input data
            amp, pha = self._get_amp_pha(data=data,
                                         which_data=which_data)
            self._amp["raw"] = amp
            self._pha["raw"] = pha
            # set background data
            self.set_bg_data(bg_data=bg_data,
                             which_data=which_data)

    def __enter__(self):
        return self

    def __eq__(self, other):
        if (np.allclose(self.amp, other.amp) and
            np.allclose(self.pha, other.pha) and
                self.meta == other.meta):
            return True
        else:
            return False

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._do_h5_cleanup:
            self.h5.flush()
            self.h5.close()

    def __repr__(self):
        rep = "QPImage, {x}x{y}px".format(x=self._amp.raw.shape[0],
                                          y=self._amp.raw.shape[1],
                                          )
        if "wavelength" in self.h5.attrs:
            wl = self.h5.attrs["wavelength"]
            if wl < 2000e-9 and wl > 10e-9:
                # convenience for light microscopy
                rep += ", λ={:.1f}nm".format(wl * 1e9)
            else:
                rep += ", λ={:.2e}m".format(wl)

        return rep

    def _get_amp_pha(self, data, which_data):
        """Convert input data to phase and amplitude

        Parameters
        ----------
        data: 2d ndarray (float or complex) or list
            The experimental data (see `which_data`)
        which_data: str
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            "field", "phase", "hologram", "phase,amplitude", or
            "phase,intensity", where the latter two require an
            indexable object with the phase data as first element.

        Returns
        -------
        amp, pha: tuple of (:py:class:`Amplitdue`, :py:class:`Phase`)
        """
        if which_data not in VALID_INPUT_DATA:
            msg = "`which_data` must be one of {}!".format(VALID_INPUT_DATA)
            raise ValueError(msg)

        if which_data == "field":
            amp = np.abs(data)
            pha = unwrap_phase(np.angle(data))
        elif which_data == "phase":
            pha = unwrap_phase(data)
            amp = np.ones_like(data)
        elif which_data == "phase,amplitude":
            amp = data[1]
            pha = unwrap_phase(data[0])
        elif which_data == "phase,intensity":
            amp = np.sqrt(data[1])
            pha = unwrap_phase(data[0])
        elif which_data == "hologram":
            amp, pha = self._get_amp_pha(holo.get_field(data),
                                         which_data="field")
        return amp, pha

    @property
    def bg_amp(self):
        """background amplitude image"""
        return self._amp.bg

    @property
    def bg_pha(self):
        """background phase image"""
        return self._pha.bg

    @property
    def amp(self):
        """background-corrected amplitude image"""
        return self._amp.image

    @property
    def field(self):
        """background-corrected complex field"""
        return self.amp * np.exp(1j * self.pha)

    @property
    def meta(self):
        """meta data"""
        return MetaDict(self.h5.attrs)

    @property
    def pha(self):
        """background-corrected phase image"""
        return self._pha.image

    @property
    def shape(self):
        """the shape of the image"""
        return self._pha.h5["raw"].shape

    def clear_bg(self, which_data=["amplitude", "phase"], keys="fit"):
        """Clear background correction

        Parameters
        ----------
        which_data: str or list of str
            From which type of data to remove the background
            information. The list contains either "amplitude",
            "phase", or both.
        keys: str or list of str
            Which type of background data to remove. One of:
              - "fit": the background data computed with
                :py:func:`qpimage.QPImage.compute_bg`
              - "data": the experimentally obtained background image

        """
        # convert to list
        if isinstance(which_data, str):
            which_data = [which_data]
        if isinstance(keys, str):
            keys = [keys]
        # Get image data for clearing
        imdats = []
        if "amplitude" in which_data:
            imdats.append(self._amp)
        if "phase" in which_data:
            imdats.append(self._pha)
        if not imdats:
            msg = "`which_data` must contain 'phase' or 'amplitude'!"
            raise ValueError(msg)
        # Perform clearing of backgrounds
        for imdat in imdats:
            for key in keys:
                imdat.set_bg(None, key)

    def compute_bg(self, which_data="phase",
                   fit_offset="average", fit_profile="ramp",
                   border_m=0, border_perc=0, border_px=0,
                   from_binary=None, ret_binary=False):
        """Compute background correction

        Parameters
        ----------
        which_data: str or list of str
            From which type of data to remove the background
            information. The list contains either "amplitude",
            "phase", or both.
        fit_profile: str
            The type of background profile to fit:
              - "ramp": 2D linear ramp with offset (default)
              - "offset": offset only
        fit_offset: str
            The method for computing the profile offset
              - "fit": offset as fitting parameter
              - "gauss": center of a gaussian fit
              - "mean": simple average
              - "mode": mode (see `qpimage.bg_estimate.mode`)
        border_m: float
            Assume that a frame of `border_m` meters around the
            image is background. The value is converted to
            pixels and rounded.
        border_perc: float
            Assume that a frame of `border_perc` percent around
            the image is background. The value is converted to
            pixels and rounded. If the aspect ratio of the image
            is not one, then the average of the data's shape is
            used to compute the percentage in pixels.
        border_px: float
            Assume that a frame of `border_px` pixels around
            the image is background.
        from_binary: boolean np.ndarray or None
            Use a boolean array to define the background area.
            The binary image must have the same shape as the
            input data.
        ret_binary: bool
            Return the binary image used to compute the background.

        Notes
        -----
        The `border_*` values are translated to pixel values and
        the largest pixel border is used to generate a binary
        image for background computation.

        If any of the `border_*` arguments are non-zero and
        `from_binary` is given, the intersection of the two
        resulting binary images is used.

        See Also
        --------
        qpimage.bg_estimate.estimate
        """
        # convert to list
        if isinstance(which_data, str):
            which_data = [which_data]
        # check validity
        if not ("amplitude" in which_data or
                "phase" in which_data):
            msg = "`which_data` must contain 'phase' or 'amplitude'!"
            raise ValueError(msg)
        # get border in px
        border_list = []
        if border_m:
            if border_m < 0:
                raise ValueError("`border_m` must be greater than zero!")
            border_list.append(border_m / self.meta["pixel size"])
        if border_perc:
            if border_perc < 0 or border_perc > 50:
                raise ValueError("`border_perc` must be in interval [0, 50]!")
            size = np.average(self.shape)
            border_list.append(size * border_perc / 100)
        if border_px:
            border_list.append(border_px)
        # get maximum border size
        if border_list:
            border_px = np.int(np.round(np.max(border_list)))
        elif from_binary is None:
            raise ValueError("Neither `from_binary` nor `border_*` given!")
        elif np.all(from_binary == 0):
            raise ValueError("`from_binary` must not be all-zero!")
        # Get affected image data
        imdat_list = []
        if "amplitude" in which_data:
            imdat_list.append(self._amp)
        if "phase" in which_data:
            imdat_list.append(self._pha)
        # Perform correction
        for imdat in imdat_list:
            binary = imdat.estimate_bg(fit_offset=fit_offset,
                                       fit_profile=fit_profile,
                                       border_px=border_px,
                                       from_binary=from_binary,
                                       ret_binary=ret_binary)
        return binary

    def copy(self, h5file=None):
        """Create a copy of the current instance

        This is done by recursively copying the underlying hdf5 data.

        Parameters
        ----------
        h5file: str, h5py.File, h5py.Group, or None
            see `QPImage.__init__`
        """
        h5 = copyh5(self.h5, h5file)
        return QPImage(h5file=h5)

    def refocus(self, distance, method="helmholtz"):
        """Numerically refocus the current field

        Parameters
        ----------
        distance: float
            Focusing distance [m]
        method: str
            Refocusing method, one of ["helmholtz","fresnel"]

        See Also
        --------
        :mod:`nrefocus` library used for numerical focusing
        """
        # TODO:
        # - Perform refocusing and create new image data instances
        # - Remember old image data instances
        # - Maybe return a new instance of QPImage
        # - Allow autofocusing?
        raise NotImplementedError("refocusing not implemented")

    def set_bg_data(self, bg_data, which_data=None):
        """Set background amplitude and phase data

        Parameters
        ----------
        bg_data: 2d ndarray (float or complex), list, QPImage, or `None`
            The background data (must be same type as `data`).
            If set to `None`, the background data is reset.
        which_data: str
            String or comma-separated list of strings indicating
            the order and type of input data. Valid values are
            "field", "phase", "phase,amplitude", or "phase,intensity",
            where the latter two require an indexable object with
            the phase data as first element.
        """
        if isinstance(bg_data, QPImage):
            if which_data is not None:
                msg = "`which_data` must not be set if `bg_data` is QPImage!"
                raise ValueError(msg)
            pha, amp = bg_data.pha, bg_data.amp
        elif bg_data is None:
            # Reset phase and amplitude
            amp, pha = None, None
        else:
            # Compute phase and amplitude from data and which_data
            amp, pha = self._get_amp_pha(bg_data, which_data)
        # Set background data
        self._amp.set_bg(amp, key="data")
        self._pha.set_bg(pha, key="data")


def copyh5(inh5, outh5):
    """Recursively copy all hdf5 data from one group to another

    Parameters
    ----------
    inh5: str, h5py.File, or h5py.Group
        The input hdf5 data. This can be either a file name or
        an hdf5 object.
    outh5: str, h5py.File, h5py.Group, or None
        The output hdf5 data. This can be either a file name or
        an hdf5 object. If set to `None`, a new hdf5 object is
        created in memory.

    Notes
    -----
    All data in outh5 are overridden by the inh5 data.
    """
    if not isinstance(inh5, h5py.Group):
        inh5 = h5py.File(inh5, mode="r")
    if outh5 is None:
        # create file in memory
        h5kwargs = {"name": "qpimage{}.h5".format(QPImage._instances),
                    "driver": "core",
                    "backing_store": False,
                    "mode": "a"}
        outh5 = h5py.File(**h5kwargs)
        return_h5obj = True
        QPImage._instances += 1
    elif not isinstance(outh5, h5py.Group):
        # create new file
        outh5 = h5py.File(outh5, mode="w")
        return_h5obj = False
    else:
        return_h5obj = True
    # begin iteration
    for key in inh5:
        if key in outh5:
            del outh5[key]
        if isinstance(inh5[key], h5py.Group):
            outh5.create_group(key)
            copyh5(inh5[key], outh5[key])
        else:
            outh5[key] = copy.copy(inh5[key].value)
            outh5[key].attrs.update(inh5[key].attrs)
    outh5.attrs.update(inh5.attrs)
    if return_h5obj:
        # in-memory or previously created instance of h5py.File
        return outh5
    else:
        # properly close the file and return its name
        fn = outh5.filename
        outh5.flush()
        outh5.close()
        return fn



import abc

import numpy as np

from . import bg_estimate


class ImageData(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, h5):
        self.h5 = h5
        if "bg_data" not in self.h5:
            self.h5.create_group("bg_data")

    def __repr__(self):
        name = self.__class__.__name__
        rep = "{name} image, {x}x{y}px".format(name=name,
                                               x=self.raw.shape[0],
                                               y=self.raw.shape[1],
                                               )
        return rep

    def __setitem__(self, key, value):
        if key in self.h5:
            del self.h5[key]
        if value is not None:
            self.h5[key] = value

    @abc.abstractmethod
    def _bg_combine(self, *bgs):
        """Combine several background images"""

    @abc.abstractmethod
    def _bg_correct(self, raw, bg):
        """Remove `bg` from `raw` image data"""

    @property
    def bg(self):
        """The combined background image data"""
        return self._bg_combine(self.h5["bg_data"].values())

    @property
    def image(self):
        """The background corrected image data"""
        return self._bg_correct(self.raw, self.bg)

    @property
    def raw(self):
        return self.h5["raw"].value

    def estimate_bg(self, fit_offset="average", fit_profile="ramp",
                    border_px=0, from_binary=None, ret_binary=False):
        """Estimate image background

        Parameters
        ----------
        fit_profile: str
            The type of background profile to fit:
              - "ramp": 2D linear ramp with offset (default)
              - "offset": offset only
        fit_offset: str
            The method for computing the profile offset
              - "fit": offset as fitting parameter
              - "gauss": center of a gaussian fit
              - "mean": simple average
              - "mode": mode (see `qpimage.bg_estimate.mode`)
        border_px: float
            Assume that a frame of `border_px` pixels around
            the image is background.
        from_binary: boolean np.ndarray or None
            Use a boolean array to define the background area.
            The binary image must have the same shape as the
            input data.
        ret_binary: bool
            Return the binary image used to compute the background.

        Notes
        -----
        If both `border_px` and `from_binary` are given, the
        intersection of the two resulting binary images is used.

        The arguments passed to this method are stored in the
        hdf5 file `self.h5` and are used for optional integrity
        checking using `qpimage.integrity_check.check`.

        See Also
        --------
        qpimage.bg_estimate.estimate
        """
        # remove existing bg before accessing imdat.image
        self.set_bg(bg=None, key="fit")
        # compute bg
        bgimage, binary = bg_estimate.estimate(data=self.image,
                                               fit_offset=fit_offset,
                                               fit_profile=fit_profile,
                                               border_px=border_px,
                                               from_binary=from_binary,
                                               ret_binary=True)
        attrs = {"fit_offset": fit_offset,
                 "fit_profile": fit_profile,
                 "border_px": border_px}
        self.set_bg(bg=bgimage, key="fit", attrs=attrs)
        # save `from_binary` separately (arrays vs. h5 attributes)
        # (if `from_binary` is `None`, this will remove the array)
        self["estimate_bg_from_binary"] = from_binary
        # return binary image
        if ret_binary:
            return binary

    def get_bg(self, key=None, ret_attrs=False):
        """Get the background data

        Parameters
        ----------
        key: None or str
            A user-defined key that identifies the background data.
            Examples are "data" for experimental data, or "fit"
            for an estimated background correction. If set to `None`,
            returns the combined background image (`ImageData.bg`).
        ret_attrs: bool
            Also returns the attributes of the background data.
        """
        if key is None:
            if ret_attrs:
                raise ValueError("No attributes for combined background!")
            return self.bg
        else:
            if key in self.h5["bg_data"]:
                data = self.h5["bg_data"][key].value
                attrs = self.h5["bg_data"][key].attrs
                if ret_attrs:
                    ret = (data, attrs)
                else:
                    ret = data
            else:
                raise KeyError("No background data for {}!".format(key))
        return ret

    def set_bg(self, bg, key="data", attrs={}):
        """Set the background data

        Parameters
        ----------
        bg: int, float, 2d ndarray, or same subclass of ImageData
            The background data. If set to `None`, the data will be
            removed.
        key: str
            A user-defined key that identifies the background data.
            Examples are "data" for experimental data, or "ramp"
            for a ramp background correction. There are no
            restrictions regarding key names.
        attrs: dict
            List of background attributes
        """
        # remove previous background key
        if key in self.h5["bg_data"]:
            del self.h5["bg_data"][key]
        # set background
        if bg is not None:
            msg = "`bg` must be scalar or ndarray"
            assert isinstance(bg, (float, int, np.ndarray)), msg
            self.h5["bg_data"][key] = bg
            for kw in attrs:
                self.h5["bg_data"][key].attrs[kw] = attrs[kw]


class Amplitude(ImageData):
    def _bg_combine(self, bgs):
        """Combine several background amplitude images"""
        out = np.ones(self.h5["raw"].shape, dtype=float)
        # Use indexing ([:]), because bg is an h5py.DataSet
        for bg in bgs:
            out *= bg.value
        return out

    def _bg_correct(self, raw, bg):
        """Remove background from raw amplitude image"""
        return raw / bg


class Phase(ImageData):
    def _bg_combine(self, bgs):
        """Combine several background phase images"""
        out = np.zeros(self.h5["raw"].shape, dtype=float)
        for bg in bgs:
            # Use .value attribute, because bg is an h5py.DataSet
            out += bg.value
        return out

    def _bg_correct(self, raw, bg):
        """Remove background from raw phase image"""
        return raw - bg



from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402


def test_clear_bg():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    bg_amp = np.linspace(.93, 1.02, size**2).reshape(size, size)
    bg_pha = pha * .5

    qpi = qpimage.QPImage(data=(pha, amp),
                          bg_data=(bg_pha, bg_amp),
                          which_data="phase,amplitude")
    # amplitude
    assert np.all(qpi.bg_amp == bg_amp)
    qpi.clear_bg(which_data="amplitude", keys="data")
    assert np.allclose(qpi.bg_amp, 1)
    qpi.compute_bg(which_data="amplitude",
                   fit_offset="fit",
                   fit_profile="ramp",
                   border_px=5)
    assert not np.allclose(qpi.bg_amp, 1)
    qpi.clear_bg(which_data="amplitude", keys="fit")
    assert np.allclose(qpi.bg_amp, 1)
    # phase
    assert np.all(qpi.bg_pha == bg_pha)
    qpi.clear_bg(which_data="phase", keys="data")


def test_clear_bg_error():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    qpi = qpimage.QPImage(pha, which_data="phase")
    try:
        qpi.clear_bg(which_data="gretel")
    except ValueError:
        pass
    else:
        assert False, "clear bg of invalid which_data should not work"


def test_comput_bg_error():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    qpi = qpimage.QPImage(pha, which_data="phase",
                          meta_data={"pixel size": .1})
    try:
        qpi.compute_bg(which_data="hänsel", border_px=4)
    except ValueError:
        pass
    else:
        assert False, "invalid which_data used for compute_bg"
    try:
        qpi.compute_bg(which_data="phase")
    except ValueError:
        pass
    else:
        assert False, "computation of bg from undefined binary"
    try:
        qpi.compute_bg(which_data="phase", border_m=-1)
    except ValueError:
        pass
    else:
        assert False, "negative border not allowed"
    try:
        qpi.compute_bg(which_data="phase", border_perc=60)
    except ValueError:
        pass
    else:
        assert False, "more than 50 percent border not allowed"
    try:
        qpi.compute_bg(which_data="phase", border_perc=-10)
    except ValueError:
        pass
    else:
        assert False, "negative percent border not allowed"


def test_get_amp_pha():
    # make sure no errors when printing repr
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)
    field = amp * np.exp(1j * pha)
    inten = amp**2

    qpi1 = qpimage.QPImage(pha, which_data="phase")
    qpi2 = qpimage.QPImage([pha, amp], which_data="phase,amplitude")
    qpi3 = qpimage.QPImage([pha, inten], which_data="phase,intensity")
    qpi4 = qpimage.QPImage(field, which_data="field")

    # test phases
    assert np.allclose(qpi1.pha, qpi2.pha)
    assert np.allclose(qpi1.pha, qpi3.pha)
    assert np.allclose(qpi1.pha, qpi4.pha)

    # test amplitudes
    assert np.allclose(qpi1.amp, 1)
    assert np.allclose(qpi2.amp, qpi3.amp)
    assert np.allclose(qpi2.amp, qpi4.amp)


def test_get_amp_pha_invalid():
    size = 50
    data = np.zeros((size, size))
    try:
        qpimage.QPImage(data, which_data="jean-luc")
    except ValueError:
        pass
    else:
        assert False, "invalid which_data should not work"


def test_get_amp_pha_holo():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    image = np.sin(kx * x + ky * y)
    qpi = qpimage.QPImage(image, which_data="hologram")
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="ramp",
                   border_px=5)
    assert np.abs(qpi.pha).max() < .1


def test_properties():
    size = 50
    pha = np.repeat(np.linspace(0, 10, size), size)
    pha = pha.reshape(size, size)
    amp = np.linspace(.95, 1.05, size**2).reshape(size, size)

    bg_amp = np.linspace(.93, 1.02, size**2).reshape(size, size)
    bg_pha = pha * .5

    qpi = qpimage.QPImage((pha, amp), bg_data=(bg_pha, bg_amp),
                          which_data="phase,amplitude")

    assert np.all(qpi.bg_amp == bg_amp)
    assert np.all(qpi.bg_pha == bg_pha)

    assert np.iscomplexobj(qpi.field)
    fval = -0.46418608511978926 + 0.91376822116221712j
    assert np.allclose(qpi.field[20, 20], fval)

    assert qpi.shape == (size, size)

    qpi.compute_bg(which_data=["phase", "amplitude"],
                   fit_offset="fit",
                   fit_profile="ramp",
                   border_px=5)

    assert not np.all(qpi.bg_amp == bg_amp)
    assert not np.all(qpi.bg_pha == bg_pha)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402


def test_qpimage_holo():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    disk_max = 1.5
    # there is a phase disk as data in the hologram
    data = disk_max * ((x - size / 2)**2 + (y - size / 2)**2 < 30**2)
    image = np.sin(kx * x + ky * y + data)
    qpi = qpimage.QPImage(image, which_data="hologram")
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="ramp",
                   border_px=5)
    assert np.allclose(disk_max, qpi.pha.max(), rtol=.01, atol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpimage  # noqa: E402


def test_repr():
    # make sure no errors when printing repr
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    qpi = qpimage.QPImage(phase, which_data="phase",
                          meta_data={"wavelength": 550e-9})
    print(qpi)
    qpi2 = qpimage.QPImage(phase, which_data="phase",
                           meta_data={"wavelength": .1})
    print(qpi2)

    print(qpi._pha)
    print(qpi._amp)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
